#pragma once

#include <cstdint>

namespace BGTU::ComputerGraphicsLabWork::RendererApi {
	class Color {
	public:
		using component_fast_t = std::uint_fast8_t;
		using component_compact_t = std::uint_least8_t;

		static constexpr component_fast_t component_min_value = 0;
		static constexpr component_fast_t component_max_value = 255;

		component_compact_t red;
		component_compact_t green;
		component_compact_t blue;

		constexpr Color() noexcept: red{0}, green{0}, blue{0} {}

		constexpr Color(component_fast_t red, component_fast_t green, component_fast_t blue) noexcept: red{red}, green{green}, blue{blue} {}

		constexpr bool operator==(Color const &other) const noexcept {
			return this->red == other.red && this->green == other.green && this->blue == other.blue;
		}

		constexpr bool operator!=(Color const &other) const noexcept {
			return !(*this == other);
		}


		class Transparent {
		public:
			component_compact_t red;
			component_compact_t green;
			component_compact_t blue;
			component_compact_t alpha;

			constexpr Transparent() noexcept: red{component_min_value}, green{component_min_value}, blue{component_min_value}, alpha{component_max_value} {}

			constexpr Transparent(component_fast_t red, component_fast_t green, component_fast_t blue) noexcept: red{red}, green{green}, blue{blue}, alpha{component_max_value} {}

			constexpr Transparent(component_fast_t red, component_fast_t green, component_fast_t blue, component_fast_t alpha) noexcept: red{red}, green{green}, blue{blue}, alpha{alpha} {}

			constexpr Transparent(Color c) noexcept: red{c.red}, green{c.green}, blue{c.blue}, alpha{component_max_value} {}

			constexpr bool operator==(Color::Transparent const &other) const noexcept {
				return this->red == other.red && this->green == other.green && this->blue == other.blue && this->alpha == other.alpha;
			}

			constexpr bool operator!=(Color::Transparent const &other) const noexcept {
				return !(*this == other);
			}

			[[nodiscard]] constexpr Color without_alpha() const noexcept {
				return Color{this->red, this->green, this->blue};
			};
		};
	};
}