#pragma once

#include "painter.hpp"
#include "drawing_assets.hpp"

class BarycentricInterpolator : public Shader {
private:
    float x0, y0, x1, y1, x2, y2, S;
    COLOR C0, C1, C2;
public:
    inline BarycentricInterpolator(float _x0, float _y0, float _x1, float _y1, float _x2, float _y2, COLOR A0, COLOR A1, COLOR A2) :
            x0(_x0), y0(_y0), x1(_x1), y1(_y1), x2(_x2), y2(_y2),
            S((_y1 - _y2) * (_x0 - _x2) + (_x2 - _x1) * (_y0 - _y2)),
            C0(A0), C1(A1), C2(A2) {}

    inline BarycentricInterpolator(Point p0, Point p1, Point p2, COLOR A0, COLOR A1, COLOR A2) :
            BarycentricInterpolator(p0.x, p0.y, p1.x, p1.y, p2.x, p2.y, A0, A1, A2) {}

    COLOR get_color(int x, int y) const override;
};


class RadialBrush : public Shader {
private:
    float cx, cy;
    COLOR C0, C1;
    float angle;
public:
    RadialBrush(float _x0, float _y0, float _x1, float _y1, COLOR A0, COLOR A1, float _angle) :
            cx((_x0 + _x1) / 2.0f), cy((_y0 + _y1) / 2.0f),
            C0(A0), C1(A1), angle(_angle) {}


    COLOR get_color(int x, int y) const override;
};