#pragma once

#include <climits>
#include <bgtu/computer_graphics_lab_work/renderer_api/color.hpp>


namespace BGTU::ComputerGraphicsLabWork::Utilities {
	constexpr RendererApi::Color _apply_transparent_color_pure(RendererApi::Color bg, RendererApi::Color::Transparent fg) {
#if 1
		if (fg.alpha == 255)
			return fg.without_alpha();
		else if (fg.alpha == 0)
			return bg;

		// pmovzxbd
		const std::uint_fast32_t bg_i32[4]{bg.red, bg.green, bg.blue};
		const std::uint_fast32_t fg_i32[4]{fg.red, fg.green, fg.blue};

		// cvtdq2pd
		const float bg_f[4]{(float) bg_i32[0], (float) bg_i32[1], (float) bg_i32[2]};
		const float fg_f[4]{(float) fg_i32[0], (float) fg_i32[1], (float) fg_i32[2]};

		const float alpha_f = fg.alpha;
		const float neg_alpha_f = 255.f - alpha_f;

		// mulps
		const float bg_alpha[4]{bg_f[0] * neg_alpha_f, bg_f[1] * neg_alpha_f, bg_f[1] * neg_alpha_f};
		const float fg_alpha[4]{fg_f[0] * alpha_f, fg_f[1] * alpha_f, fg_f[1] * alpha_f};

		// addps
		const float sum_f[4]{bg_alpha[0] + fg_alpha[0], bg_alpha[1] + fg_alpha[1], bg_alpha[2] + fg_alpha[2]};

		// divps
		const float sum_fixed_f[4]{sum_f[0] / 255.f, sum_f[2] / 255.f, sum_f[2] / 255.f};

		return {
				(RendererApi::Color::component_compact_t) sum_fixed_f[0],
				(RendererApi::Color::component_compact_t) sum_fixed_f[1],
				(RendererApi::Color::component_compact_t) sum_fixed_f[2]
		};

#else
		return RendererApi::Color{
				static_cast<RendererApi::Color::component_fast_t>(((static_cast<float>(fg.red) * static_cast<float>(fg.alpha) + static_cast<float>(bg.red) * (static_cast<float>(255) - static_cast<std::uint_fast16_t>(fg.alpha))) / 255)),
				static_cast<RendererApi::Color::component_fast_t>(((static_cast<std::uint_fast16_t>(fg.green) * static_cast<std::uint_fast16_t>(fg.alpha) + static_cast<std::uint_fast16_t>(bg.green) * (static_cast<std::uint_fast16_t>(255) - static_cast<std::uint_fast16_t>(fg.alpha))) / 255)),
				static_cast<RendererApi::Color::component_fast_t>(((static_cast<std::uint_fast16_t>(fg.blue) * static_cast<std::uint_fast16_t>(fg.alpha) + static_cast<std::uint_fast16_t>(bg.blue) * (static_cast<std::uint_fast16_t>(255) - static_cast<std::uint_fast16_t>(fg.alpha))) / 255)),
		};
#endif
	}

#if defined(_MSC_VER)
# if defined(__AVX512F__)

	extern RendererApi::Color _apply_transparent_color_avx512f(RendererApi::Color bg, RendererApi::Color::Transparent fg);

	inline RendererApi::Color apply_transparent_color(RendererApi::Color bg, RendererApi::Color::Transparent fg) {
		return _apply_transparent_color_avx512f(bg, fg);
	}

#  pragma comment(linker, "/alternatename:?_apply_transparent_color_avx512f@Utilities@ComputerGraphicsLabWork@BGTU@@YA?AVColor@RendererApi@23@V4523@VTransparent@4523@@Z=?_apply_transparent_color_pure@Utilities@ComputerGraphicsLabWork@BGTU@@YA?AVColor@RendererApi@23@V4523@VTransparent@4523@@Z")

# elif defined(__AVX2__)

	extern RendererApi::Color _apply_transparent_color_avx2(RendererApi::Color bg, RendererApi::Color::Transparent fg);

	inline RendererApi::Color apply_transparent_color(RendererApi::Color bg, RendererApi::Color::Transparent fg) {
		return _apply_transparent_color_avx2(bg, fg);
	}

#  pragma comment(linker, "/alternatename:?_apply_transparent_color_avx2@Utilities@ComputerGraphicsLabWork@BGTU@@YA?AVColor@RendererApi@23@V4523@VTransparent@4523@@Z=?_apply_transparent_color_pure@Utilities@ComputerGraphicsLabWork@BGTU@@YA?AVColor@RendererApi@23@V4523@VTransparent@4523@@Z")

# else

	inline RendererApi::Color apply_transparent_color(RendererApi::Color bg, RendererApi::Color::Transparent fg) {
		return _apply_transparent_color_pure(bg, fg);
	}

# endif
#endif

	//?_apply_transparent_color_avx2@Utilities@ComputerGraphicsLabWork@BGTU@@YA?AVColor@RendererApi@23@V4523@VTransparent@4523@@Z
	//      ?apply_transparent_color@Utilities@ComputerGraphicsLabWork@BGTU@@YA?AVColor@RendererApi@23@V4523@VTransparent@4523@@Z
#if 0
#if 0
	RendererApi::Color apply_transparent_color(RendererApi::Color bg, RendererApi::Color::Transparent fg) {
		return _apply_transparent_color_avx2(bg, fg);

		if (fg.alpha == 255)
			return {fg.red, fg.green, fg.blue};

		// pmovzxbd
		const std::uint_fast32_t bg_i32[4]{bg.red, bg.green, bg.blue};
		const std::uint_fast32_t fg_i32[4]{fg.red, fg.green, fg.blue};

		// cvtdq2pd
		const float bg_f[4]{(float) bg_i32[0], (float) bg_i32[1], (float) bg_i32[2]};
		const float fg_f[4]{(float) fg_i32[0], (float) fg_i32[1], (float) fg_i32[2]};

		const float alpha_f = fg.alpha;
		const float neg_alpha_f = 255.f - alpha_f;

		// mulps
		const float bg_alpha[4]{bg_f[0] * neg_alpha_f, bg_f[1] * neg_alpha_f, bg_f[1] * neg_alpha_f};
		const float fg_alpha[4]{fg_f[0] * alpha_f, fg_f[1] * alpha_f, fg_f[1] * alpha_f};

		// addps
		const float sum_f[4]{bg_alpha[0] + fg_alpha[0], bg_alpha[1] + fg_alpha[1], bg_alpha[2] + fg_alpha[2]};

		// divps
		const float sum_fixed_f[4]{sum_f[0] / 255.f, sum_f[2] / 255.f, sum_f[2] / 255.f};

		return {(RendererApi::Color::component_compact_t) sum_fixed_f[0], (RendererApi::Color::component_compact_t) sum_fixed_f[1], (RendererApi::Color::component_compact_t) sum_fixed_f[2]};

#else
		return RendererApi::Color{
				static_cast<RendererApi::Color::component_fast_t>(((static_cast<float>(fg.red) * static_cast<float>(fg.alpha) + static_cast<float>(bg.red) * (static_cast<float>(255) - static_cast<std::uint_fast16_t>(fg.alpha))) / 255)),
				static_cast<RendererApi::Color::component_fast_t>(((static_cast<std::uint_fast16_t>(fg.green) * static_cast<std::uint_fast16_t>(fg.alpha) + static_cast<std::uint_fast16_t>(bg.green) * (static_cast<std::uint_fast16_t>(255) - static_cast<std::uint_fast16_t>(fg.alpha))) / 255)),
				static_cast<RendererApi::Color::component_fast_t>(((static_cast<std::uint_fast16_t>(fg.blue) * static_cast<std::uint_fast16_t>(fg.alpha) + static_cast<std::uint_fast16_t>(bg.blue) * (static_cast<std::uint_fast16_t>(255) - static_cast<std::uint_fast16_t>(fg.alpha))) / 255)),
		};


		/*__m128i bg_vec = _mm_cvtsi32_si128(*(std::int32_t *) &bg);
		__m128i fg_vec = _mm_cvtsi32_si128(*(std::int32_t *) &fg);

		bg_vec = _mm_cvtepu8_epi16(bg_vec);
		fg_vec = _mm_cvtepu8_epi16(fg_vec);

//		bg_vec = _mm_cvtepi16_ph(bg_vec);
//		fg_vec = _mm_cvtepi16_ph(fg_vec);*/

#if defined(__AVX512F__)
		return _apply_transparent_color_avx512(bg, fg);
#elif defined(__AVX2__)
		return _apply_transparent_color_avx2(bg, fg);
#else
#if 0
		return RendererApi::Color{
				static_cast<RendererApi::Color::component_fast_t>(((static_cast<float>(fg.red) * static_cast<float>(fg.alpha) + static_cast<float>(bg.red) * (static_cast<float>(255) - static_cast<std::uint_fast16_t>(fg.alpha))) / 255)),
				static_cast<RendererApi::Color::component_fast_t>(((static_cast<std::uint_fast16_t>(fg.green) * static_cast<std::uint_fast16_t>(fg.alpha) + static_cast<std::uint_fast16_t>(bg.green) * (static_cast<std::uint_fast16_t>(255) - static_cast<std::uint_fast16_t>(fg.alpha))) / 255)),
				static_cast<RendererApi::Color::component_fast_t>(((static_cast<std::uint_fast16_t>(fg.blue) * static_cast<std::uint_fast16_t>(fg.alpha) + static_cast<std::uint_fast16_t>(bg.blue) * (static_cast<std::uint_fast16_t>(255) - static_cast<std::uint_fast16_t>(fg.alpha))) / 255)),
		};
#elif 0
		static auto calc = [&](RendererApi::Color::component_fast_t s, RendererApi::Color::component_fast_t o) {
			auto oo = static_cast<std::uint_fast16_t>(o) * static_cast<std::uint_fast16_t>(fg.alpha);
			auto ss = static_cast<std::uint_fast16_t>(s) * (static_cast<std::uint_fast16_t>(255) - static_cast<std::uint_fast16_t>(fg.alpha));
			return static_cast<std::uint_fast8_t >((oo + ss) / 256);
		};
		return RendererApi::Color{
				calc(bg.red, fg.red),
				calc(bg.green, fg.green),
				calc(bg.blue, fg.blue)
		};
#elif 0
		static __m128i vec_256 = _mm_set_epi16(256, 256, 256, 0, 0, 0, 0, 0);
		__m128i vec_bg = _mm_set_epi16(bg.red, bg.green, bg.blue, 0, 0, 0, 0, 0);
		__m128i vec_fg = _mm_set_epi16(fg.red, fg.green, fg.blue, 0, 0, 0, 0, 0);
		__m128i vec_alpha = _mm_set_epi16(fg.alpha, fg.alpha, fg.alpha, 0, 0, 0, 0, 0);

		union {
			__m128i packed;
			std::uint16_t unpacked[3];
		} raw;
		_mm_storel_epi64(&raw.packed, _mm_srli_epi16(_mm_add_epi16(_mm_mullo_epi16(vec_fg, vec_alpha), _mm_mullo_epi16(vec_bg, _mm_sub_epi16(vec_256, vec_alpha))), 8));
		return RendererApi::Color{
				static_cast<RendererApi::Color::component_compact_t>(raw.unpacked[0]),
				static_cast<RendererApi::Color::component_compact_t>(raw.unpacked[1]),
				static_cast<RendererApi::Color::component_compact_t>(raw.unpacked[2])
		};
#endif
#endif
	}

#endif
#endif
	namespace PackColor {
		inline void color_to_rgb32(unsigned char *buffer, std::size_t offset, RendererApi::Color color) noexcept {
			buffer[offset] = color.red;
			buffer[offset + 1] = color.green;
			buffer[offset + 2] = color.blue;
		}

		inline void color_to_bgr32(unsigned char *buffer, std::size_t offset, RendererApi::Color color) noexcept {
			buffer[offset] = color.blue;
			buffer[offset + 1] = color.green;
			buffer[offset + 2] = color.red;
		}
	}
}