#pragma once

#include <cstdlib>
#include <type_traits>
#include <utility>
#include <limits>
#include <bgtu/computer_graphics_lab_work/renderer_api/point.hpp>
#include "line.hpp"

namespace BGTU::ComputerGraphicsLabWork::Utilities::Shapes {

	template<class receiver_t>
	void iterate_triangle_fill(
			RendererApi::PointI<2>::component_t x0, RendererApi::PointI<2>::component_t y0,
			RendererApi::PointI<2>::component_t x1, RendererApi::PointI<2>::component_t y1,
			RendererApi::PointI<2>::component_t x2, RendererApi::PointI<2>::component_t y2,
			receiver_t receiver
	);

	class triangle_fill_iterable {
	private:
		brezenham_line_iterable upper_left;
		brezenham_line_iterable lower_left;
		brezenham_line_iterable right;

		template<class receiver_t>
		friend
		void iterate_triangle_fill(
				RendererApi::PointI<2>::component_t x0, RendererApi::PointI<2>::component_t y0,
				RendererApi::PointI<2>::component_t x1, RendererApi::PointI<2>::component_t y1,
				RendererApi::PointI<2>::component_t x2, RendererApi::PointI<2>::component_t y2,
				receiver_t
		);


		inline triangle_fill_iterable(
				brezenham_line_iterable upper_left,
				brezenham_line_iterable lower_left,
				brezenham_line_iterable right
		) : upper_left{upper_left}, lower_left{lower_left}, right{right} {}

		struct _pre_constructor {
		private:
			inline static void sort_points(RendererApi::PointI<2>::component_t *x0, RendererApi::PointI<2>::component_t *y0, RendererApi::PointI<2>::component_t *x1, RendererApi::PointI<2>::component_t *y1, RendererApi::PointI<2>::component_t *x2, RendererApi::PointI<2>::component_t *y2) {
				auto _swap = [](RendererApi::PointI<2>::component_t *sx0, RendererApi::PointI<2>::component_t *sy0, RendererApi::PointI<2>::component_t *sx1, RendererApi::PointI<2>::component_t *sy1) {
					RendererApi::PointI<2>::component_t tmp;

					tmp = *sx0;
					*sx0 = *sx1;
					*sx1 = tmp;

					tmp = *sy0;
					*sy0 = *sy1;
					*sy1 = tmp;
				};

				if (*y0 < *y1)
					_swap(x0, y0, x1, y1);
				if (*y1 < *y2)
					_swap(x1, y1, x2, y2);
				if (*y0 < *y1)
					_swap(x0, y0, x1, y1);
				if (*y0 == *y1 && *x0 < *x1)
					_swap(x0, y0, x1, y1);
				if (*y1 == *y2 && *x2 < *x1)
					_swap(x1, y1, x2, y2);
			}

		public:

			static inline triangle_fill_iterable create(
					RendererApi::PointI<2>::component_t x0, RendererApi::PointI<2>::component_t y0,
					RendererApi::PointI<2>::component_t x1, RendererApi::PointI<2>::component_t y1,
					RendererApi::PointI<2>::component_t x2, RendererApi::PointI<2>::component_t y2
			) {
				_pre_constructor::sort_points(&x0, &y0, &x1, &y1, &x2, &y2);
				return triangle_fill_iterable{brezenham_line_iterable{x0, y0, x1, y1}, brezenham_line_iterable{x1, y1, x2, y2}, brezenham_line_iterable{x0, y0, x2, y2}};
			}
		};

	public:
		inline triangle_fill_iterable(
				RendererApi::PointI<2>::component_t x0, RendererApi::PointI<2>::component_t y0,
				RendererApi::PointI<2>::component_t x1, RendererApi::PointI<2>::component_t y1,
				RendererApi::PointI<2>::component_t x2, RendererApi::PointI<2>::component_t y2
		) : triangle_fill_iterable{_pre_constructor::create(x0, y0, x1, y1, x2, y2)} {
		}

#if 0
		class iterator {
	private:
		brezenham_line_iterable::iterator upper_left;
		brezenham_line_iterable::iterator lower_left;
		brezenham_line_iterable::iterator right;
		RendererApi::PointI<2>::component_t cx, ex, cy;

		friend class triangle_fill;

		inline iterator(
				brezenham_line_iterable const &upper_left,
				brezenham_line_iterable const &lower_left,
				brezenham_line_iterable const &right
		) : upper_left{upper_left.begin()}, lower_left{lower_left.begin()}, right{right.begin()}, cx{(*this->upper_left).x}, ex{(*this->right).x}, cy{(*this->upper_left).y} {}


		brezenham_line_iterable::iterator &_left() {
			if (this->upper_left == nullptr)
				return this->lower_left;
			else
				return this->upper_left;
		}

	public:
		inline iterator &operator++() {
			if (this->cx < this->ex) {
				this->cx++;
				return *this;
			}
			if (this->upper_left != nullptr) {
				RendererApi::PointI<2>::component_t sx = (*(this->upper_left)).x;
				while (true) {
					this->upper_left++;
					if (this->upper_left == nullptr)
						goto LOWER;
					if ((*(this->upper_left)).y < this->cy)
					{
						this->cy--;
						return *this;

					(*++(this->upper_left)).y >= this->cy
					if ((*(this->upper_left)).x < sx)
						return *this;
				}
			} else {
				LOWER:
				RendererApi::PointI<2>::component_t sx = (*(this->upper_left)).x;
			}

			while ((*++(this->right)).y >= this->cy) {
				if ((*(this->right)).x > this->ex)
					return *this;
			}

			this->cy--;


		}

		inline iterator operator++(int) {
			iterator next = *this;
			++(*this);
			return next;
		}

		inline point operator*() const {
			if (this->swap_flag)
				return point{this->ret, this->arg};
			else
				return point{this->arg, this->ret};
		}

		inline bool operator==(std::nullptr_t) const {
			return this->arg >= this->end_arg;
		}

		inline bool operator!=(std::nullptr_t) const {
			return this->arg < this->end_arg;
		}
	};
#endif
	};


	template<class receiver_t>
	void iterate_triangle_fill(
			RendererApi::PointI<2>::component_t x0, RendererApi::PointI<2>::component_t y0,
			RendererApi::PointI<2>::component_t x1, RendererApi::PointI<2>::component_t y1,
			RendererApi::PointI<2>::component_t x2, RendererApi::PointI<2>::component_t y2,
			receiver_t receiver
	) {
		triangle_fill_iterable init{x0, y0, x1, y1, x2, y2};
		auto right = init.right.begin();
		RendererApi::PointI<2>::component_t cy = (*right).y;
		RendererApi::PointI<2>::component_t sx = std::numeric_limits<RendererApi::PointI<2>::component_t>::min();
		RendererApi::PointI<2>::component_t ex = std::numeric_limits<RendererApi::PointI<2>::component_t>::max();

		auto loop = [&](brezenham_line_iterable left) {

			for (auto lp: left) {
				if (lp.y == cy) {
					receiver(true, lp.x, lp.y);
					sx = (lp.x > sx) ? (lp.x) : sx;
					continue;
				}

				for (; (*right).y > lp.y; right++) {
					if ((*right).x > sx)
						receiver(true, (*right).x, (*right).y);
					ex = ((*right).x < ex) ? ((*right).x) : ex;
				}
				for (RendererApi::PointI<2>::component_t x = sx + 1; x < ex; x++)
					receiver(false, x, cy);

				cy--;
				receiver(true, lp.x, lp.y);
			}
		};

		loop(init.upper_left);
		loop(init.lower_left);
		for (; right != nullptr; right++) {
			if ((*right).x > sx)
				receiver(true, (*right).x, (*right).y);
		}
	}

	template<class receiver_t>
	void iterate_triangle_fill(
			RendererApi::PointI<2> p0,
			RendererApi::PointI<2> p1,
			RendererApi::PointI<2> p2,
			receiver_t receiver
	) {
		iterate_triangle_fill(
				p0.x, p0.y,
				p1.x, p1.y,
				p2.x, p2.y,
				receiver
		);
	}

	template<class variable_t>
	struct triangle_barycentric_interpolator {
	public:
		struct pole {
			RendererApi::PointI<2> coord;
			variable_t value;
		};
	private:
		long double full_square;
		RendererApi::PointI<2> p0;
		RendererApi::PointI<2> p1;
		RendererApi::PointI<2> p2;
		variable_t v0;
		variable_t v1;
		variable_t v2;

		struct _empty {
		};

		std::conditional_t<std::is_same_v<variable_t, RendererApi::Color::Transparent>, bool, _empty> skip_alpha;

		static long double calculate_square_2(
				RendererApi::PointI<2> p0,
				RendererApi::PointI<2> p1,
				RendererApi::PointI<2> p2
		) noexcept {
			return (p2.x - p1.x) * (p2.y - p0.y) + (p2.x - p0.x) * (p1.y - p0.y);
		}


	public:
		triangle_barycentric_interpolator(
				RendererApi::PointI<2> p0, RendererApi::PointI<2> p1, RendererApi::PointI<2> p2,
				variable_t v0, variable_t v1, variable_t v2
		) : p0{p0}, p1{p1}, p2{p2}, v0{v0}, v1{v1}, v2{v2}, full_square{calculate_square_2(p0, p1, p2)} {
			if constexpr (std::is_same_v<variable_t, RendererApi::Color::Transparent>) {
				this->skip_alpha = v0.alpha == v1.alpha && v0.alpha == v2.alpha;
			}
		}

		variable_t interpolate_point(RendererApi::PointI<2> p) const {
			const double k0 = calculate_square_2(p, this->p1, this->p2) / this->full_square;
			const double k1 = calculate_square_2(p, this->p2, this->p0) / this->full_square;
			const double k2 = calculate_square_2(p, this->p0, this->p1) / this->full_square;
			auto f = [&]<class intermediate_t>(intermediate_t (variable_t::*offset)) -> intermediate_t {
				return static_cast<intermediate_t>((this->v0.*offset) * k0 + (this->v1.*offset) * k1 + (this->v2.*offset) * k2);
			};
			if constexpr (std::is_same_v<variable_t, RendererApi::Color::Transparent>) {
				RendererApi::Color base{
						f(&RendererApi::Color::Transparent::red),
						f(&RendererApi::Color::Transparent::green),
						f(&RendererApi::Color::Transparent::blue),
				};
				return RendererApi::Color::Transparent{
						base.red, base.green, base.blue,
						(this->skip_alpha ? this->v0.alpha : f(&RendererApi::Color::Transparent::alpha))
				};
			} else if constexpr (std::is_same_v<variable_t, RendererApi::Color>) {
				return RendererApi::Color{
						f(&RendererApi::Color::red),
						f(&RendererApi::Color::green),
						f(&RendererApi::Color::blue)
				};
			} else {
				return static_cast<variable_t>(f(&variable_t::variable_t));
			}
		}

		variable_t interpolate_point(RendererApi::PointI<2>::component_t x, RendererApi::PointI<2>::component_t y) const {
			return this->interpolate_point(RendererApi::PointI<2>{x, y});
		}
	};


	class triangle_barycentric_shader : public Shader {
	private:
		triangle_barycentric_interpolator<RendererApi::Color::Transparent> interpolator;
	public:
		inline triangle_barycentric_shader(
				RendererApi::PointI<2> p0, RendererApi::PointI<2> p1, RendererApi::PointI<2> p2,
				RendererApi::Color::Transparent c0, RendererApi::Color::Transparent c1, RendererApi::Color::Transparent c2
		) : interpolator{p0, p1, p2, c0, c1, c2} {}


		[[nodiscard]] inline RendererApi::Color::Transparent get_color(RendererApi::PointI<2>::component_t x, RendererApi::PointI<2>::component_t y) const final {
			return this->interpolator.interpolate_point(x, y);
		}
	};
}